// To compile on Windows

// Install GLFW 3.3.8
// https://www.glfw.org/download.html
// Install GLEW 2.2.0
// https://github.com/nigels-com/glew/releases/tag/glew-2.2.0
//
// extract the downloaded .zip files to "C:/libs"; this is currently expected
// by our CMakeLists.txt.

// Install CMake
// https://cmake.org/download
// Add to PATH for all users
// from project root:
// mkdir build
// cd build
// cmake ..
// cmake --build .
// The last step compiles the executable - this can also be done from Visual
// Studio

// To run in VS
// Set startup project in Solution Explorer
// Press F5 to run

// To run in VSCode
// https://code.visualstudio.com/docs/cpp/config-mingw

// To compile on Arch Linux
//
// Install dependencies
// sudo pacman -S glfw mesa glew
//
// Build
// cmake ..
// cmake --build .

#include <GL/glew.h>
#include <GLFW/glfw3.h>
#include <iostream>

#include "gfx.hpp"
#include "icosphere.hpp"
#include "orbiter.hpp"
#include "orbitervisualizer.hpp"
#include "orbitvisualizer.hpp"
#include "widget.hpp"

#include <skein/orbit.h>

// INPUT!
//
// what input do we even want in the first place?
// * camera controls - this is a rendering only concern which needn't affect the orbital
// * model validation - we want to modify orbits over time to confirm that our model is
// working properly
//
// TODO: input can only be directly handled by static methods, so we need to find a way to collect
// input from object instances and handle it at a higher level. in the case of this class, we want
// to determine when a configurable key is pressed. alternatively, we can split the behaviour into
// multiple sub-classes and run different loops at the top level?
//
// another idea is to process all input into a well-defined Input struct in the main loop, then
// pass this struct into objects' render() method.
//
// it's possible the first idea makes the most sense if the plan is to ultimately extract the orbit
// stuff to a library, since it keeps input a separate concern from the physics model
struct Input
{
    bool cycleAnimation;
} input;

void keyCallback(GLFWwindow* window, int key, int scancode, int action, int mods)
{
    if (key == GLFW_KEY_C && action == GLFW_PRESS)
    {
        input.cycleAnimation = true;
    }
}

void clearInput()
{
    input.cycleAnimation = false;
}

int main()
{
    GLFWwindow* window = nullptr;
    if (initGraphics(&window, "skeingl") != 0)
        return -1;

    GLuint litProgram = compileShaderProgram("./frag_lit.glsl");
    GLuint unlitProgram = compileShaderProgram("./frag_unlit.glsl");

    // set up scene
    Icosphere planet(0.2, 3, litProgram);

    //Orbiter orbiter(orbiterSphere, orbit, unlitProgram);
    Orbiter orbiter;
    Orbit& orbit = orbiter.getOrbit();
    orbit.setSemiMajorAxis(.75);
    orbit.setEccentricity(.5);
    orbit.setInclination(3.142 / 2.0 + 1);
    orbit.setArgumentOfPeriapsis(2.0);
    orbit.setLongitudeOfAscendingNode(0.1);

    Icosphere orbiterSphere(0.07, 2, litProgram);
    OrbiterVisualizer orbiterVisualizer(orbiter, orbiterSphere, unlitProgram);
    OrbitVisualizer orbitVisualizer(orbit, unlitProgram);

    // register input
    glfwSetKeyCallback(window, keyCallback);

    // TODO: convert these to an enum
    const int ANIM_ORBITING     = 0;
    const int ANIM_ECCENTRICITY = 1;
    int animation = 0;
    float time = glfwGetTime();

    // Main loop
    while (!glfwWindowShouldClose(window))
    {
        // receive input - key callback populates input struct defined further up,
        // clearInput() needs to be called to clear input from previous frame
        clearInput();
        glfwPollEvents();

        // apply input
        if (input.cycleAnimation)
        {
            animation++;
            if (animation > ANIM_ECCENTRICITY)
            {
                animation = 0;
            }
        }

        // only update time if playing the orbiting animation
        if (animation == ANIM_ORBITING)
        {
            time = glfwGetTime();
        }
        else
        {
            float e = .25 + .2 * sin(glfwGetTime());
            orbit.setEccentricity(e);
        }

        // rendering
        glClearColor(0.2, 0.3, 0.3, 1.0);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        planet.render(time);
        orbiterVisualizer.render(time);
        orbitVisualizer.render(time);

        glfwSwapBuffers(window);
    }

    glfwTerminate();
    return 0;
}
